///////////////////////////////////////////////////////////////
//                                                           //
//  cmpStrs.cpp - adds logical comparisons to strs           //
//  v1.1                                                     //
//                implementations                            //
//                                                           //
//  Language:     Borland C++                                //
//  System:       Gateway 386/25, MS-DOS 6.0                 //
//  Application:  CSE 687                                    //
//  Author:       Jim Fawcett, Syracuse University           //
//                (315)443-2654, jfawcett@cat.syr.edu        //
//                                                           //
//                    copyright (c) 1996                     //
//    all rights granted to users except right to publish    //
//                                                           //
///////////////////////////////////////////////////////////////

#include <string.h>     // strlen()
#include "strs.hpp"
#include "cmpStrs.hpp"


//----< comparison >-------------------------------------------

int cmpStr::operator==(const str &s) const {

  if(strcmp(cStr(),s.cStr())) return 0; else return 1;
}

//----< comparison >-------------------------------------------

int cmpStr::operator==(const char *s) const {

  if(strcmp(cStr(),(char*)s)) return 0; else return 1;
}

//----< comparison >-------------------------------------------

int cmpStr::operator<=(const str &s) const {

  if(strcmp(cStr(),s.cStr()) > 0) return 0; else return 1;
}

//----< comparison >-------------------------------------------

int cmpStr::operator<=(const char *s) const {

  if(strcmp(cStr(),s) > 0) return 0; else return 1;
}

//----< comparison >-------------------------------------------

int cmpStr::operator< (const str &s) const {

  if(strcmp(cStr(),s.cStr()) >= 0) return 0; else return 1;
}

//----< comparison >-------------------------------------------

int cmpStr::operator< (const char *s) const {

  if(strcmp(cStr(),s) >= 0) return 0; else return 1;
}


//

//----< test stub >--------------------------------------------

#ifdef TEST_CMPSTRS

#include <fstream.h>

//----< test function accepting a cmpStr object >--------------

void testFunc(cmpStr s) { cout << "  " << s << endl; }

//----< tests start here >-------------------------------------

void main() {

  cout << "\n"
       << "  =====================================\n"
       << "   Testing comparisons of text strings \n"
       << "  =====================================\n\n";

  cout << "\n"
       << " test #1 - void constr, appending, stream insertion:\n"
       << "-----------------------------------------------------\n";
  cmpStr testStr1;
  testStr1.showState();
  int i;
  for(i=0; i<20; i++)
    testStr1.append(char(int('a')+i));
  cout << "  testStr1 = " << testStr1 << endl;
  testStr1.showState();

  cout << "\n"
       << " test #2 - copy constructor:\n"
       << "-----------------------------\n";
  cmpStr testStr2 = testStr1;
  cout << "  testStr2 = " << testStr2 << endl;

  cout << "\n"
       << " test #3 - C string promotion construction:\n"
       << "--------------------------------------------\n";
  cmpStr testStr3 = "platypus";
  cout << "  testStr3 = " << testStr3 << endl;

  cout << "\n"
       << " test #4 - assignment from cmpStr:\n"
       << "-----------------------------------\n";
  testStr3 = testStr2;
  cout << "  testStr3 = " << testStr3 << endl;
  testStr3.showState();  cout << endl;

  cout << "\n"
       << " test #5 - assignment from C string:\n"
       << "-------------------------------------\n";

  const int numCreatures = 5;
  char* test[numCreatures] = {
    "ardvarrk", "viper", "platypus", "moose", "badger"
  };
  testStr3 = test[1];
  cout << "  testStr3 = " << testStr3 << endl;
  testStr3.showState();  cout << endl;

//

  cout << "\n"
       << " test #6 - truncation:\n"
       << "-----------------------\n";
  testStr3.len(3);
  cout << "  testStr3 = " << testStr3 << endl;
  testStr3.showState();
  cout << endl;

  cout << "\f\n\n"
       << " test #7 - concatenation with operator+=():\n"
       << "--------------------------------------------\n";
  testStr1.flush();
  for(i=0; i<numCreatures; i++) {
    testStr1+=cmpStr(test[i]);
    testStr1+=" ";
  }
  cout << "  " << testStr1 << endl;

  cout << "\n"
       << " test #8 - implicit and explicit coercions:\n"
       << "--------------------------------------------\n";
  testFunc("the rain in Spain falls mainly on the plain");
  testFunc(cmpStr("the rain in Spain falls mainly on the plain"));

  cout << "\n"
       << " test #9 - array of strings:\n"
       << "-----------------------------\n";

  cmpStr array[4] = { "now", "is", "the", "hour" };
  for(i=0; i<4; i++)
    cout << "  array[" << i << "] = " << array[i] << endl;

  cout << "\n"
       << " test #10 - heap memory allocation:\n"
       << "------------------------------------\n";

  cmpStr *onHeap = new cmpStr("stored on heap");
  cout << "  *onHeap = " << *onHeap << endl;

//

  cout << "\n"
       << " test #11 - stack memory allocation:\n"
       << "-------------------------------------\n";

  for(i=0; i<3; i++) {
    cout << "  bytes allocated at start = " << cmpStr::alloc() << endl;
    cmpStr s = "the rain in Spain falls mainly on the plain";
    cout << "  "  << s << ": bytes allocated = " << s.alloc()  << endl;
    {
      cmpStr s = "Jack be nimble, Jack be quick";
      cout << "  " << s << ": bytes allocated = " << s.alloc() << endl;
    }
    cout << "  bytes allocated = " << cmpStr::alloc()          << endl;
  }
  cout << "  bytes allocated at end   = " << cmpStr::alloc()   << endl;

  cout << "\n"
       << " test #12 - word extraction from file:\n"
       << "---------------------------------------\n";

  ifstream in("strs.cpp",ios::in);
  if(!in.good())
    cout << "\n  can't open file strs.cpp\n";
  else
    cout << "\n  first 10 lines of file strs.cpp are:\n\n";
  cmpStr inStr;
  for(i=0; i<10; i++) {
    in >> inStr;
    cout << "  " << inStr << endl;
  }
  cout << endl;

  cout << "\f\n #13 - testing comparisons:\n"
       <<     "----------------------------\n";

  cmpStr s1, s2;
  char *s3;
  for(i=0; i<numCreatures; i++) {
    s1 = test[i];
    s2 = test[3];
    if(s2> s1) cout << "  " << s2 << " >  " << s1 << endl;
    if(s2>=s1) cout << "  " << s2 << " >= " << s1 << endl;
    if(s2==s1) cout << "  " << s2 << " == " << s1 << endl;
    if(s2!=s1) cout << "  " << s2 << " != " << s1 << endl;
    if(s2<=s1) cout << "  " << s2 << " <= " << s1 << endl;
    if(s2< s1) cout << "  " << s2 << " <  " << s1 << endl;
  }
  for(i=0; i<numCreatures; i++) {
    s1 = test[4];
    s3 = test[i];
    if(s1> s3) cout << "  " << s1 << " >  " << s3 << endl;
    if(s1>=s3) cout << "  " << s1 << " >= " << s3 << endl;
    if(s1==s3) cout << "  " << s1 << " == " << s3 << endl;
    if(s1!=s3) cout << "  " << s1 << " != " << s3 << endl;
    if(s1<=s3) cout << "  " << s1 << " <= " << s3 << endl;
    if(s1< s3) cout << "  " << s1 << " <  " << s3 << endl;
  }

}

#endif
