#ifndef CMPSTRS_HPP
#define CMPSTRS_HPP
///////////////////////////////////////////////////////////////
//                                                           //
//  cmpStrs.hpp - adds logical comparisons to strs           //
//  v1.1                                                     //
//                manual page                                //
//                                                           //
//  Language:     Borland C++                                //
//  Platform:     Gateway 386/25                             //
//                MS-DOS 6.2, win 95.400.950                 //
//  Application:  CSE 687                                    //
//  Author:       Jim Fawcett, Syracuse University           //
//                (315)443-2654, jfawcett@cat.syr.edu        //
//                                                           //
//                    copyright (c) 1997                     //
//    all rights granted to users except right to publish    //
//                                                           //
///////////////////////////////////////////////////////////////
/*

    cmpStr class
    ============
    This class provides facilities for making lexographic com-
    parisons between str objects.  It is endowed with all the
    operations expected of str objects.  However, it can also
    be used as a mixin with other classes derived from the str
    class.  That is, a client can use the cmpStr class by form-
    ing a mixin with other classes derived from strs.

    constructors:
      cmpStr s;                       construct with 10 char alloc
      cmpStr s(15);                   construct with 15 char alloc
      cmpStr s2 = s1;                 construct by copying s1
      cmpStr s = "this is a string";  construct from C string

    testing string state ( >, >=, ==, !=, <=, < ):
      s1 == s2   s1 == "a string"     lexographic equality
      s1 != s2   s1 != "a string"     lexographic inequality
      s1 <  s2   s1 <  "a string";    lexographic order
      s1 <= s2   s1 <= "a string";    lexographic order
      s1 >= s2   s1 >= "a string";    lexographic order
      s1 >  s2   s1 >  "a string";    lexographic order

    other operations:
      All operations described on the strs module manual page
      are available for cmpStr objects since they are public-
      ally derived from the str class.
*/
//

///////////////////////////////////////////////////////////////
//                                                           //
//  Build Process:                                           //
//                                                           //
//  Files required for building test stub:                   //
//      strs.hpp, strs.cpp, cmpStrs.hpp, cmpStrs.cpp         //
//                                                           //
//  Build from integrated environment:                       //
//  - open project citing the *.cpp files above              //
//  - set Options\Compiler\Code generated\defines to         //
//      TEST_CMPSTRS                                         //
//  - run                                                    //
//                                                           //
//  - command line build, no memory allocation test:         //
//      bcc -w+ -DTEST_CMPSTRS strs.cpp cmpStrs.cpp          //
//                                                           //
///////////////////////////////////////////////////////////////
/*

   maintenance history:
   ====================
   - v1.1 : 15 Jan 1997
     . fixed bug in void allocation due to redefinition of a default
       argument.  Fix required adding an explicit void constructor.
   - v1.0 : 19 Jan 1997
     . created cmpStrs module to serve as mixin with strs module



*/
//
///////////////////////////////////////////////////////////////
//                                                           //
//                      declarations                         //
//                                                           //
///////////////////////////////////////////////////////////////


#include "strs.hpp"

class cmpStr : public virtual str {

  public:
    cmpStr( )               : str(10) { }
    cmpStr(const cmpStr &c) : str(c) { }
    cmpStr(const   char *s) : str(s) { }
    cmpStr(int size)        : str(size) { }

    int operator==(const str &s) const;
    int operator!=(const str &s) const;
    int operator< (const str &s) const;
    int operator<=(const str &s) const;
    int operator>=(const str &s) const;
    int operator> (const str &s) const;

    int operator==(const char *s) const;
    int operator!=(const char *s) const;
    int operator< (const char *s) const;
    int operator<=(const char *s) const;
    int operator>=(const char *s) const;
    int operator> (const char *s) const;

};


inline int cmpStr::operator!=(const str &s) const {
  return !operator==(s);
}

inline int cmpStr::operator!=(const char *s) const {
  return !operator==(s);
}

inline int cmpStr::operator>(const str &s) const {
  return !operator<=(s);
}

inline int cmpStr::operator>(const char *s) const {
  return !operator<=(s);
}

inline int cmpStr::operator>=(const str &s) const {
  return !operator< (s);
}

inline int cmpStr::operator>=(const char *s) const {
  return !operator< (s);
}

//
///////////////////////////////////////////////////////////////
//                                                           //
//  Design Notes                                             //
//                                                           //
///////////////////////////////////////////////////////////////
/*

   This class provides facilities for making lexographic compari-
   sons between str objects.  It can be used as a mixin with other
   classes derived from the str class.  That is, a client can use
   cmpStr class "as-is" or by forming a mixin with other classes
   derived from strs to get more functionality, e.g.:

     class myStr : public virtual cmpStr, manStr {

       public:
         myStr(int s=10)       : str(s);
         myStr(const  char *s) : str(s);
         myStr(const myStr &s) : str(s);
     };

     myStr ms = "this is a string";

   This construction allows both lexographic comparisons and manip-
   ulation operations.  The inheritance is virtual so that only one
   copy of str class object is contained within each mixin object.

                           +-------+
                           |  str  |
                           +---+---+
                               |
                              /_\
                               |
                    +----------+----------+
                    |                     |
               +----+----+           +----+----+
               |  manStr |           |  cmpStr |
               +----+----+           +----+----+
                    |                     |
                    +----------+----------+
                               |
                              /_\
                               |
                           +---+---+
                           | myStr |
                           +-------+

   manStr is publically derived from str so it inherits all str
   operations.


*/
#endif

