#ifndef MAP_HPP
#define MAP_HPP


///////////////////////////////////////////////////////////////
//                                                           //
//  map.hpp    keep track key-value associations in the map  //
//  v1.0                                                     //
//                manual page                                //
//                                                           //
//  Language:     Borland C++,Version 3.1                    //
//  Platform:                                                //
//                MS-DOS 6.2, Windows 95 v4.400.950          //
//  Application:  CSE 687  Pr2 shall                         //
//  Author:       Mehmet SEN, CIS Grad Student at SU         //
//                (315) 443-2142, msen@top.cis.syr.edu       //
//                                                           //
//                    copyright (c) 1996                     //
//    all rights granted to users except right to publish    //
//                                                           //
///////////////////////////////////////////////////////////////
//    information about the class operations
/*
	Map class
	===========================================
	Store any key-value associations in the sorted order in a
	map and provides some useful manipulations on the map.
	Provides accesses to the associations like arrays with
	the below mentioned extended operations.
	Keys and values can be any type, provided keys with
	comparison capabilities by themselves.

	constructors
	===========================================
	map m;			construct m with 10 places
	map m1(50);		construct m1 with 50 places
	map m2(m); 		copy constructor
	map m3(key1,val1 [,40]); construct with initial key value

	operations
	===========================================
	m1=m2			assignment operator
	m.insert(key,val)       insert key-value association to the map
	m.lookUp(key)		return to key associated value,like [] oper.
	m.has(key,index)	check existence of key, (index is the
				place of the key in the map)
	m[key]=val=m[key2]      access to a specific key value
	checkKey(index)		returns the key at the specific location in the
				map for advanced usages
	quickValue		get value by index without searching
	length			returns the number of associations
	dump			display the current map
	search			search for a specific key, returns best match
	nearMatch 		perform nearMatch with   substring
				properties
	noMatch			in case of nothing found

	hidden
	======
	openMap			construct a number of associations
	enlarge			increase the size
	binsearch(key,start,end)   find a key in the map, BINARY search
	defaultVal		produce a global value, NECESSARY
				to avoid errors, which may occur
				by the misuse of the [] operator.
	abort			no more space

*/

//
///////////////////////////////////////////////////////////////
//                                                           //
//  Build Process:                                           //
//                                                           //
//  Files required for building test stub: (on strings)      //
//      map.hpp, map.cpp, cmpStrs.hpp, cmpStrs.cpp           //
//                        strs.hpp, strs.cpp                 //
//  Build from integrated environment:                       //
//  - close any open project                                 //
//  - set Options\Compiler\Code generated\defines to         //
//    if you want to perform general tests on string classes //
//      TEST_MAP                                             //
//    if you want to test it to list the files in current dir//
//  - open map.cpp in edit window and run                    //
//                                                           //
//  - command line build,                                    //
//      bcc -w+ -DTEST_MAP map.cpp cmpstrs.cpp               //
//                                 strs.cpp                  //
//      OR for the file list				     //
//      bcc -w+ -DTEST_MAP_FILES map.cpp cmpstrs.cpp         //
//                                 strs.cpp                  //
//  - P.S. if both TEST_MAP_FILES and TEST_MAP 	are defined  //
//          TEST_MAP_FILES hide TEST_MAP definition          //
//                                                           //
///////////////////////////////////////////////////////////////
/*


	maintenance history
	===========================================
	- v1.1 :        March 1,1997
			Addition of file list tests
	- v1.0 :	February 19,1997
			-first release
	design notes
	===========================================

	Map class keeps keys and values in two arrays, which both have
	same indexes for the specified associations.

	No duplication of the keys are allowed; duplications
	are not correlated with the sprit of the map. (However,if you need
	them, you will see the usage of the map in a different way
	in the tree module. Map class provides useful functions for
	making virtual duplicates, like checkKey(),  see tree module.)



	Key and value objects are from template classes, can be any
	object. However, the comparision operators for the keys, and
	<< operators for both of them must be defined to dump the map.
*/
//
///////////////////////////////////////////////////////////////               
//                                                           //               
//                      declarations                         //               
//                                                           //               
///////////////////////////////////////////////////////////////               

#include <iostream.h>
#include <stdlib.h>
#include "cmpstrs.hpp"           // for the testing with the string classes
#include "strs.hpp"


//----< map class >----------------------------------------

template <class key, class val>
class map {
  public:
    map(int sz=10);                             //constructors
    map(const map & m);
    map(const key &k,const val &v, int sz=10);
    map<key,val> & operator=(const map<key,val> &m); //assign.
    void insert(const key &k,const val &v);     // operations
    val & lookUp(const key &k);
    val & operator[](const key &k);
    const key & checkKey(int index) const;
    val & quickValue(int index) const;
    int has(const key &k,int *index);
    void dump(ostream & out);
    int length() const;
    ~map();                                    //destructor

    val & search(const key &k,int *i,int *points);
    virtual val & nearMatch(const key &k,int *i,int *points);
    virtual void noMatch(){}

  protected:                                     //hidden operations
    val & defaultVal();
    void openMap();
    int binsearch(const key & k,int start,int end);


    void enlarge();
    void abort(const char *s);

    val defaultValue;                         //default value for assignment
    key * KEYS;                               //key value arrays
    val  *VALS;
    int size;                                 //physcial length
    int howmany;                              //logical length
    int enlargement;                          //size increment amount
};


//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// BASIC FUNCTIONS: 3 constructurs, assignment, Destructor,
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/

/////////////////////////////////////////////////////////////
//  MAP Constructor
/////////////////////////////////////////////////////////////
template <class key, class val>
map<key,val>::map(int sz=10)
		:size(sz), howmany(0), enlargement(sz) 	 {
	openMap();     //allocate size for key-values
}

/////////////////////////////////////////////////////////////
//  MAP constructor with initial values
/////////////////////////////////////////////////////////////
template <class key, class val>
map<key,val>::map(const key& k, const val & v,int sz=10)
		:size(sz), howmany(0), enlargement(sz) 	 {
	openMap();     //allocate size for key-values
	insert(k,v);   //insert first element
}

/////////////////////////////////////////////////////////////
//  MAP COPY Constructor
/////////////////////////////////////////////////////////////
template <class key, class val>
map<key,val>::map(const map<key,val> &m) {

	size=m.size;
	howmany=m.howmany;
	enlargement=m.enlargement;
	openMap();
	for (int i=0;i<howmany;i++){
		KEYS[i]=m.KEYS[i];
		VALS[i]=m.VALS[i];
	}
}
/////////////////////////////////////////////////////////////
//  MAP ASSIGNMENT OPERATOR
/////////////////////////////////////////////////////////////
template <class key, class val>
map<key,val> & map<key,val>::operator=(const map<key,val> &m) {
	if (this==&m) return *this;
	size=m.size;
	howmany=m.howmany;
	enlargement=m.enlargement;
	openMap();
	for (int i=0;i<howmany;i++){
		KEYS[i]=m.KEYS[i];
		VALS[i]=m.VALS[i];
	}
	return *this;
}
/////////////////////////////////////////////////////////////
// destructor, deallocate spaces for the arrays
/////////////////////////////////////////////////////////////
template <class key, class val>
map<key,val>::~map() {
	delete [] KEYS;
	delete [] VALS;
}

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// SIDE SERVICE FUNCTIONS; defaults , abort , space allocation
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/

/////////////////////////////////////////////////////////////
// Produce a global default value to prevent errors by using [] operator
/////////////////////////////////////////////////////////////
template <class key, class val>
val & map<key,val>::defaultVal(){
	static val default_val; //Note that it is necessary to define global, also see
	return default_val;		// page 286 of textbook
}

/////////////////////////////////////////////////////////////
// Terminate the program with the error message
/////////////////////////////////////////////////////////////
template <class key, class val>
void  map<key,val>::abort(const char *s){
	cout <<s<<endl;
	exit(1);
}

/////////////////////////////////////////////////////////////
// Make space allocations for a new map
/////////////////////////////////////////////////////////////
template <class key, class val>
void map<key,val>::openMap(){
	defaultValue=defaultVal();
	KEYS=new key[size];        // space allocation for the keys
	if (!KEYS) abort("Memory Allocation Failure.");
	VALS=new val[size];        // space allocation for the values
	if (!VALS) abort("Memory Allocation Failure!");
}

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// MAIN SERVICE FUNCTIONS
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/

/////////////////////////////////////////////////////////////
//  return the logical length of the map
/////////////////////////////////////////////////////////////
template <class key, class val>
int map<key,val>::length() const {
     return howmany;
}
/////////////////////////////////////////////////////////////
//  print everything in the map to an output stream
/////////////////////////////////////////////////////////////
template <class key, class val>
void map<key,val>::dump(ostream & out) {
	out << "Dumping the map:"<<endl;
	for (int i=0; i<howmany ;i++){
		out << KEYS[i] << "-->" <<VALS[i]<<endl;
	}
}

/////////////////////////////////////////////////////////////
// Insert a new association to the map and make proper changes
/////////////////////////////////////////////////////////////
template <class key, class val>
void map<key,val>::insert(const key &_k,const val &_v){
	int i;
	key k(_k);		// make a copy of the objects
	val v(_v);
	if ( has(k,&i) ) {          // do not allow duplication
		(*this)[k]=v;
		return;
	}
	if (howmany==size)          // no more space, increment the space
		enlarge();
	for (i=howmany-1; i>=0 ; i--){     //insertion sort
		if ( k<KEYS[i]) {
			KEYS[i+1]=KEYS[i];
			VALS[i+1]=VALS[i];  //this is not an actual object
		}				// move
		else 	break;
	}
	i++; howmany++;
	KEYS[i]=k;          // insert element
	VALS[i]=v;
}

/////////////////////////////////////////////////////////////
//  Enlarge the present map space by the amount of enlargement value
/////////////////////////////////////////////////////////////
template <class key, class val>
void map<key,val>::enlarge(){
	size +=enlargement;
	key * tempKEYS=new key[size];
	val * tempVALS=new val[size];
	for (int i=0;i<howmany;i++){
		tempKEYS[i]=KEYS[i];
		tempVALS[i]=VALS[i];
	}
	delete [] KEYS;
	delete [] VALS;
	KEYS=tempKEYS;
	VALS=tempVALS;
}

/////////////////////////////////////////////////////////////
// Search the map for a specific key; BINARY SEARCH
/////////////////////////////////////////////////////////////
template <class key, class val>
map<key,val>::binsearch(const key &k,int start,int end)
{
	int mid=(start+end)/2;
	if (start>end) return -1 ;  //not found
	else if (KEYS[mid]==k) {return mid;}
	else if ( KEYS[mid]>k ) { return binsearch(k,start,mid-1);}
	return (binsearch(k,mid+1,end));
}
/////////////////////////////////////////////////////////////
// Search the map for a specific key; BINARY SEARCH
/////////////////////////////////////////////////////////////
template <class key, class val>
val & map<key,val>::search(const key &k,int * index,int * points)
{
	*index=binsearch(k,0,howmany-1);
	*points=0;
	if ( -1 != *index ) {
		*points=0;
		return quickValue(*index) ;
	} else {	             // BINARY SEARCH CAN NOT FOUND EXACT
		val &v=nearMatch(k,index,points);// index, and points for the near one
		if (*index != -1)
			return v;	  // near match found
		else 	noMatch();       // near match not found
	}
	return defaultValue;       // nothing found
}
/////////////////////////////////////////////////////////////
// Return the value for a specific KEY
/////////////////////////////////////////////////////////////
template <class key, class val>
val &  map<key,val>::lookUp(const key &k){
	int index,p;
	return search(k,&index,&p);
}

/////////////////////////////////////////////////////////////
// Check whether key exists or not
/////////////////////////////////////////////////////////////
template <class key, class val>
int map<key,val>::has(const key &k, int * index)
{
	int percent;
	search(k,index,&percent);
	if (-1==*index)
		return 0;
	return 1;
}

/////////////////////////////////////////////////////////////
// Operator [] for the map usage in the programs
/////////////////////////////////////////////////////////////
template <class key, class val>
val & map<key,val>::operator[](const key &k)
{
    return lookUp(k);
}

/////////////////////////////////////////////////////////////
// get key at a specific index,(useful  for the making virtual duplicates
/////////////////////////////////////////////////////////////
template <class key, class val>
const key & map<key,val>::checkKey(int index) const
{
	    return KEYS[index];
}
/////////////////////////////////////////////////////////////
// get association value by the index of the asssociation
/////////////////////////////////////////////////////////////
template <class key, class val>
val & map<key,val>::quickValue(int index) const
{
	    return VALS[index];
}
/////////////////////////////////////////////////////////////
// return the best matched key value, i.e., exact match, else default val.
/////////////////////////////////////////////////////////////
template <class key, class val>
val & map<key,val>::nearMatch(const key &k,int *i,int *points)
{       if ( k==k)
		*i=*points=-1;      // no search operation performed
	return defaultValue;
}
#endif //MAP_HPP
