#ifndef MAP1_HPP
#define MAP1_HPP
///////////////////////////////////////////////////////////////
//                                                           //
//  map1.hpp    stores key-value associations in the map     //
//  v1.0         with special access functions, near match   //
//                manual page                                //
//                                                           //
//  Language:     Borland C++,Version 3.1                    //
//  Platform:                                                //
//                MS-DOS 6.2, Windows 95 v4.400.950          //
//  Application:  CSE 687  Pr3 tutor                         //
//  Author:       Mehmet SEN, CIS Grad Student at SU         //
//                (315) 443-2142, msen@top.cis.syr.edu       //
//                                                           //
//                    copyright (c) 1996                     //
//    all rights granted to users except right to publish    //
//                                                           //
///////////////////////////////////////////////////////////////
//    information about the class operations
/*
	Map1 class
	===========================================
	Map1 class is derived from the map class;
	Store any key-value associations in the sorted order,ascending,
	in a map and provides some useful manipulations on the map.
	Provides accesses to the associations like arrays with
	the below mentioned extended operations.
	Keys and values can be any type, provided keys with
	comparison capabilities by themselves.

	Map1 class provides near match capabilities while the clients
	are accessing to the map associations. Near matches are based
	upon substring cases without case sensitivity to provide
	flexibility.

	constructors
	===========================================
	map m1;			construct m with 10 places
	map m1(50);		construct m1 with 50 places
	map m2(m); 		copy constructor
	map m3(key1,val1 [,40]); construct with initial key value

	operations
	===========================================
	search			search for a specific key, returns best match
	nearMatch 		perform nearMatch with   substring
				properties
	sub			find the similarities of two word


	Other Operations
	======
      All operations described on the map module manual page
      are available for cmpStr objects since they are public-
      ally derived from the map class.

*/

//
///////////////////////////////////////////////////////////////
//                                                           //
//  Build Process:                                           //
//                                                           //
//  Files required for building test stub: (on strings)      //
//      map.hpp, map.cpp, cmpStrs.hpp, cmpStrs.cpp           //
//      map1.hpp map1,cpp strs.hpp, strs.cpp                 //
//  Build from integrated environment:                       //
//  - close any open project                                 //
//  - set Options\Compiler\Code generated\defines to         //
//    if you want to perform general tests on string classes //
//      TEST_MAP1                                            //
//    if you want to test it to list the files in current dir//
//  - open map.cpp in edit window and run                    //
//                                                           //
//  - command line build,                                    //
//      bcc -w+ -DTEST_MAP1  map1.cpp map.cpp cmpstrs.cpp    //
//                                 strs.cpp                  //
//                                                           //
///////////////////////////////////////////////////////////////
/*


	maintenance history
	===========================================
	- v1.0 :	March 18,1997
			-first release
	design notes
	===========================================
	map1 class'es nearMatch function is based on sub function
	which perform string comparisions, and return a value
	specifying the similarity of two keys.

	Future map classes can be derived from map1 class only by
	changing sub function for the different key comparisions.



*/
//
///////////////////////////////////////////////////////////////               
//                                                           //               
//                      declarations                         //               
//                                                           //               
///////////////////////////////////////////////////////////////               
#include <string.h>
#include "cmpstrs.hpp"            // for using strings in the map
#include "strs.hpp"
#include "map.hpp"		// base class


//----< map class >----------------------------------------

template <class key, class val>
class map1 :public map<key,val>{
  public:
    map1();
    map1(int sz);                             //constructors
    map1(const map<key,val> & m);
    map1(const key &k,const val &v);
    map1(const key &k,const val &v, int sz);
    map1<key,val> & operator=(const map<key,val> &m); //assign.
    ~map1();                                    //destructor

    virtual int sub(key k, key s);
    virtual val & nearMatch(const key &k,int *i,int * points);
    virtual void noMatch();

  private:                                     //hidden operations
};


//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// BASIC FUNCTIONS: 3 constructurs, assignment, Destructor,
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/

/////////////////////////////////////////////////////////////
//  MAP Constructor
/////////////////////////////////////////////////////////////
template <class key, class val>
map1<key,val>::map1() :map<key,val>() {}

template <class key, class val>
map1<key,val>::map1(int sz):map<key,val>(sz){}

/////////////////////////////////////////////////////////////
//  MAP constructor with initial values
/////////////////////////////////////////////////////////////
template <class key, class val>
map1<key,val>::map1(const key& k, const val & v,int sz):
		map<key,val>(k,v,sz){}
template <class key, class val>
map1<key,val>::map1(const key& k, const val & v):
		map<key,val>(k,v,10){}

/////////////////////////////////////////////////////////////
//  MAP COPY Constructor
/////////////////////////////////////////////////////////////
template <class key, class val>
map1<key,val>::map1(const map<key,val> &m):
		map<key,val>(m) {}

/////////////////////////////////////////////////////////////
//  MAP ASSIGNMENT OPERATOR
/////////////////////////////////////////////////////////////
template <class key, class val>
map1<key,val> & map1<key,val>::operator=(const map<key,val> &m) {
	if (this==&m) return *this;
	map<key,val>::operator=(m);
	return *this;
}
/////////////////////////////////////////////////////////////
// destructor, deallocate spaces for the arrays
/////////////////////////////////////////////////////////////
template <class key, class val>
map1<key,val>::~map1() {}
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// MAIN SERVICE FUNCTIONS
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/

/////////////////////////////////////////////////////////////
//  accepts two keys, and compare them assuming that they are
//  cmpStr classes. A similarity value returned increases when the
//  similarity decreases. Comparisions are not case sensitive to provide
//  flexibility.
/////////////////////////////////////////////////////////////
template <class key, class val>
int map1<key,val>::sub(key k,  key  s){  // objects are copied by purpose

	cmpStr t;
	int lk=k.len();
	int ls=s.len();
	if ( (lk==0) || (ls==0) ) return 30000; //impossible value
	if (lk > ls)  {
		t=s; s=k; k=t;
		ls=lk; lk=k.len();
	}
	int d=ls-lk;
	int maxSub=0,current;
	char * ck=(char *) malloc(lk+5); // need 1 more for \0 terminator
	char * cs=(char *) malloc(ls+5);  // to make it safe added 5
	strcpy(ck,k.cStr());	strlwr(ck);
	strcpy(cs,s.cStr());    strlwr(cs);
	for (int i=0; i <= d; i++) {
		current=strspn(ck,cs+i);
		if ( maxSub < current ) {
		       int ok=1;
		       for (int a=0;a < (lk*8)/10;a++)
			 if ( ck[a]!=cs[a+i] )
				ok=0;
			if (ok&&(lk>2)) maxSub =current;
		}
	}
	return (1000*(lk-maxSub))/lk + (500*(ls-maxSub))/ls +100;
}

/////////////////////////////////////////////////////////////
//  find the closest matches baes on the sub function
/////////////////////////////////////////////////////////////
template <class key, class val>
val & map1<key,val>::nearMatch(const key &k,int * i,int * points){
	//if ( sub(k,KEYS[0]) == -1 )   // means no
	int current,min=30000;
	for (int j=0; j < howmany;j++) {
		if ( (current=sub(k, KEYS[j])) < min ) {
			min=current;
			*i=j;
		}
	}
	*points=min;
	if ( min <= 800 ) return quickValue(*i) ;
	*points=-1; *i=-1;
	return defaultValue;
}

/////////////////////////////////////////////////////////////
//   in case of no match
/////////////////////////////////////////////////////////////
template <class key, class val>
void map1<key,val>::noMatch()
{
}

#endif //MAP1_HPP
