///////////////////////////////////////////////////////////////
//                                                           //
//  textwins.cpp - builds text based windows                 //
//  v3.2                                                     //
//                                                           //
///////////////////////////////////////////////////////////////
//                                                           //
//  Language:    Borland C, ver 3.1;                         //
//  Platform:    Gateway 386/25, win95                       //
//  Application: CSE687, pr3                                 //
//  Author:      Jim Fawcett, 123 Link Hall, Syracuse Univ   //
//               Syracuse, N.Y., 13244, (315)443-2654        //
//                                                           //
//           copyright (c) 1996, James W. Fawcett            //
//                all rights granted to user                 //
//                    except publication                     //
//                                                           //
///////////////////////////////////////////////////////////////

#include <fstream.h>        // ifstream()
#include <stdlib.h>         // exit()
#include <string.h>         // strlen()
#include <stdio.h>          // fprintf()
#include <ctype.h>          // isalnum()
#include <alloc.h>          // faralloc(),farfree();
#include "textwins.hpp"
#include "tmouse.hpp"


///////////////////////////////////////////////////////////////
//  key codes                                                //
///////////////////////////////////////////////////////////////

const unsigned char _curRight =  77;     // cursor right
const unsigned char _curLeft  =  75;     // cursor left
const unsigned char _ctlCurR  = 116;     // control cursor right
const unsigned char _ctlCurL  = 115;     // control cursor left
const unsigned char _curUp    =  72;     // cursor up
const unsigned char _curDn    =  80;     // cursor down
const unsigned char _ctlPgUp  = 132;     // control page up
const unsigned char _ctlPgDn  = 118;     // control page down
const unsigned char _pageUp   =  73;     // page up
const unsigned char _pageDn   =  81;     // page down
const unsigned char _end      =  79;     // go to end of line
const unsigned char _home     =  71;     // go to beginning of line
const unsigned char _backSp   =   8;     // backspace
const unsigned char _esc      =  27;     // escape
const unsigned char _F1       =  59;     // F1 function key
const unsigned char _F2       =  60;     // F2
const unsigned char _F3       =  61;     // F3
const unsigned char _F4       =  62;     // F4
const unsigned char _F5       =  63;     // F5
const unsigned char _F6       =  64;     // F6
const unsigned char _F7       =  65;     // F7
const unsigned char _F8       =  66;     // F8
const unsigned char _F9       =  67;     // F9
const unsigned char _F10      =  68;     // F10
const unsigned char _F11      =  84;     // F11
const unsigned char _F12      =  85;     // F12

//


///////////////////////////////////////////////////////////////
//  alert window default corners                             //
///////////////////////////////////////////////////////////////

const unsigned int alertLt    =  5;      // left col
const unsigned int alertTp    = 10;      // top row
const unsigned int alertWd    = 70;      // width
const unsigned int alertHt    =  5;      // height

///////////////////////////////////////////////////////////////
//  static mouse object owned by base window class           //
///////////////////////////////////////////////////////////////

  tmouse win::mouse;

///////////////////////////////////////////////////////////////
//  static flag controlling mouse event acceptance           //
///////////////////////////////////////////////////////////////

  int win::acceptAll = 0;

///////////////////////////////////////////////////////////////
//  static pointer to alert window                           //
///////////////////////////////////////////////////////////////

  win* win::alertwin=0;


///////////////////////////////////////////////////////////////
//  constants                                                //
///////////////////////////////////////////////////////////////

enum { nonInvert, Invert };


//

//----< concatenate two strings into newly alloc memory >------

char* win::concat(const char *s1, const char *s2) {

  int len1 = strlen(s1);  int len2 = strlen(s2);
  char *cat = new char[len1+len2+1];
  if(!cat) {
    cout << "memory allocation failure in msgWin::concat()\n";
    exit(1);
  }
  strcpy(cat,s1);  strcpy(&cat[len1],s2);
  return cat;
}

//----< abort with message >-----------------------------------

void win::abort(const char *s) { cout << s << endl; exit(1); }

//----< check position coordinate >----------------------------

int win::chkHor(int p) {

  if(p<1 || 80<p)
    abort("invalid horizontal coordinate");
  return p;
}

//----< check position coordinate >----------------------------

int win::chkVer(int p) {

  if(p<1 || 25<p)
    abort("invalid vertical coordinate");
  return p;
}

//----< char with foreground and background attributes >-------

unsigned int win::extchar(unsigned char ch) {

  int fg = inverseVideo ? bgc : fgc;
  int bg = inverseVideo ? fgc : bgc;
  bg = bg&7;
  return ((((bg<<4) + fg)<<8)+(ch&255));
}

//

//----< constructor >------------------------------------------

win::win(int l, int t, int w, int h) {

  lft = chkHor(l);  tp = chkVer(t);
   wd = chkHor(w);  ht = chkVer(h);
  prtSize = 0; prtBuf = 0;
  txtSize = 0; txtBuf = 0;
  textcolor(fgc=defFGC);
  textbackground(bgc=defBGC);
  inverseVideo = 0;
  bufSize = ht*wd;
  buffer = new int[bufSize+1];
  if(!buffer)
    abort("memory allocation failure in win()");
  int i;
  for(i=0; i<bufSize; i++)
    buffer[i] = extchar(' ');
  buffer[bufSize] = '\0';
  myIndex   = -1;
}

//----< copy constructor >-------------------------------------

win::win(const win &w)
    : lft(w.lft), tp(w.tp), wd(w.wd), ht(w.ht),
      bgc(w.bgc), fgc(w.fgc), inverseVideo(w.inverseVideo) {

  prtSize = w.prtSize; prtBuf = 0;
  prtBuf = new char[prtSize+1];
  if(!prtBuf)
    abort("memory allocation failure in win()");
  int i;
  for(i=0; i<prtSize; i++)
    prtBuf[i] = w.prtBuf[i];
  prtBuf[prtSize] = '\0';
  txtSize = w.txtSize; txtBuf = 0;
  txtBuf = new char[txtSize+1];
  if(!txtBuf)
    abort("memory allocation failure in win()");
  for(i=0; i<txtSize; i++)
    txtBuf[i] = w.txtBuf[i];
  txtBuf[txtSize] = '\0';
  bufSize = ht*wd;
  buffer = new int[bufSize+1];
  if(!buffer)
    abort("memory allocation failure in win()");
  for(i=0; i<bufSize; i++)
    buffer[i] = extchar(' ');
  buffer[bufSize] = '\0';
}

//

//----< set window size >--------------------------------------

win& win::size(int cols, int rows) {

  wd = chkHor(cols);  ht = chkVer(rows);
  int size = rows*cols;
  if(size > bufSize) {
    if(buffer) delete [] buffer;
    bufSize = wd*ht;  buffer = new int[bufSize+1];
    if(!buffer)
      abort("memory allocation failure in size()");
  }
  int i;
  for(i=0; i<bufSize; i++)
    buffer[i] = extchar(' ');
  buffer[bufSize] = '\0';
  return *this;
}

//----< destructor >-------------------------------------------

win::~win() {

// return buffers for window, text, and prompt
  if(buffer) { delete [] buffer; buffer = 0; }
  if(txtBuf) { delete [] txtBuf; txtBuf = 0; }
  if(prtBuf) { delete [] prtBuf; prtBuf = 0; }
}

//----< set window's position - top and left coordinates >-----

win& win::place(int x, int y) {
  lft=chkHor(x); tp=chkVer(y); return *this;
}

//----< set window prompt >------------------------------------

win& win::prompt(char *p) {

  if(prtBuf) delete [] prtBuf;
  prtBuf = new char[prtSize=strlen(p)];
  if(!prtBuf)
    abort("memory allocation failure in prompt()");
  int i;
  for(i=0; i<prtSize; i++)
    prtBuf[i] = p[i];
  return *this;
}

//

//----< return window's top row >------------------------------

int win::top() { return tp; }

//----< return window's left col >-----------------------------

int win::left() { return lft; }

//----< return width of window >-------------------------------

int win::width()  { return wd; }

//----< return height of window >------------------------------

int win::height() { return ht; }

//----< return current window row >----------------------------

int win::row() { return wherey(); }

//----< return current window column >-------------------------

int win::col() { return wherex(); }

//----< set window foreground and background colors >----------

win& win::colors(int fg, int bg) {
  fgc = fg; bgc = bg; return *this;
}

//----< set foreground color >---------------------------------

win& win::FGcolor(int c) { fgc = c; return *this; }

//----< set background color >---------------------------------

win& win::BGcolor(int c) { bgc = c; return *this; }

//----< return foreground color >------------------------------

int win::FGcolor() { return fgc; }

//----< return background color >------------------------------

int win::BGcolor() { return bgc; }

//

//----< convert window coordinates to buffer coordinates >-----

// note that window coordinates start with (1,1)

int win::map(int row, int col) {

  int index = (row-1)*wd+col-1;
  if(index<0 || bufSize <=index)
    abort("window buffer index out of range");
  return index;
}

//----< place character in window >----------------------------

win& win::charxy(unsigned char ch, int row, int col) {

  int index = map(row,col);
  if(index < 0 || bufSize <= index) {
    fprintf(stdout,"indexing error in charxy()\n");
    exit(1);
  }
  buffer[map(row,col)] = extchar(ch);
  return *this;
}

//----< place line of chars in window >------------------------

win& win::horizLine(char ch, int row, int start, int end) {

  chkVer(row); chkHor(start); chkHor(end);
  int i;
  for(i=start; i<=end; i++)
  charxy(ch,row,i);
  return *this;
}

//----< place line of chars in window >------------------------

win& win::vertiLine(char ch, int col, int start, int end) {

  chkHor(col); chkVer(start); chkVer(end);
  int i;
  for(i=start; i<=end; i++)
    charxy(ch,i,col);
  return *this;
}

//----< get text from another window >-------------------------

win& win::moveText(win &w) {

  if(this == &w) return *this;
  if(txtBuf) delete [] txtBuf;
  txtBuf = new char[txtSize=w.txtSize];
  if(!txtBuf)
    abort("memory alloc failure in moveText()");
  int i;
  for(i=0; i<txtSize; i++)
    txtBuf[i] = w.txtBuf[i];
  return *this;
}

//

//----< send text to message window >--------------------------

win& win::putText(const char *t) {

// reallocate text buffer for t
  if(!t) return *this;
  int len = strlen(t);
  if(len >=txtSize) {
    if(txtBuf) delete [] txtBuf;
    txtBuf = new char[txtSize=wd-4];
    if(!txtBuf)
      abort("memory alloc failure in putText()");
  }
// copy
  int i;
  for(i=0; i<len; i++)
    txtBuf[i] = t[i];
  for(i=len; i<txtSize; i++)
    txtBuf[i] = ' ';
  return *this;
}

//----< find last printable char in text buffer >--------------

int win::textEnd(char *buffer, int size) {

  int end = size-1;
  while(buffer[end]==' ' && end>0) end--;
  return end+1;
}

//----< return text buffer contents to client >----------------

const char* win::getText() {

  #define rbufSize 80
  static char rtnBuf[rbufSize];
  if(txtSize >=rbufSize) txtSize = rbufSize-1;
  int end   = textEnd  (txtBuf,txtSize);
  int i;
  for(i=0; i<end; i++)                      // build rtnBuf
    rtnBuf[i] = txtBuf[i];
  rtnBuf[end] = '\0';
  return rtnBuf;
  #undef rbufSize
}

//

//----< draw window borders with current focus >---------------

void win::drawBorders(int intensity) {

  window(lft,tp,lft+wd-1,tp+ht-1);
  int save = FGcolor();
  if(intensity) {
    if(bgc == LIGHTGRAY || bgc == WHITE)
      FGcolor(BLACK);
    else
      FGcolor(WHITE);
    horizLine(char(205),1,1,wd);
    horizLine(char(205),ht,1,wd);
    vertiLine(char(186),1,1,ht);
    vertiLine(char(186),wd,1,ht);
    charxy(char(201),1,1);
    charxy(char(187),1,wd);
    charxy(char(200),ht,1);
    charxy(char(188),ht,wd);
    if(prtSize) {
      int i, len = prtSize;
      if(len > wd-6) len = wd-6;
      charxy(' ',1,3);
      for(i=0; i<len; i++)
  charxy(prtBuf[i],1,i+4);
      charxy(' ',1,len+4);
    }
  }
  else {
    if(bgc == LIGHTGRAY || bgc == WHITE)
      FGcolor(CYAN);
    else
      FGcolor(YELLOW);
    horizLine(char(196),1,1,wd);
    horizLine(char(196),ht,1,wd);
    vertiLine(char(179),1,1,ht);
    vertiLine(char(179),wd,1,ht);
    charxy(char(218),1,1);
    charxy(char(191),1,wd);
    charxy(char(192),ht,1);
    charxy(char(217),ht,wd);
    if(prtSize) {
      int i, len = prtSize;
      if(len > wd-6) len = wd-6;
      charxy(' ',1,3);
      for(i=0; i<len; i++)
  charxy(prtBuf[i],1,i+4);
      charxy(' ',1,len+4);
    }
  }
  FGcolor(save);
}

//

//----< set or remove window focus >---------------------------

void win::focus(int f) { f ? drawBorders(1) : drawBorders(0); }


//----< paint message window on terminal screen >--------------

win& win::paint(int f, int invert, int justify) {

  int i,j;
  for(i=1; i<=ht; i++)
    for(j=1; j<=wd; j++)
      charxy(' ',i,j);
  if(txtSize) {
    if(invert) toggleVideo();
    int len = textEnd(txtBuf,txtSize);
    if(len > wd-2) len = wd-2;
    int row = (ht+1)>>1;
    int textStart;
    if(ht > 0) {
      if(justify) {
        textStart = wd-len+2;
        if(textStart > 3)
          textStart >>= 1;
        else
          textStart = 2;
      }
      else
        textStart = 3;
      if(ht == 1)
        for(i=2; i<wd; i++)
          charxy(' ',row,i);
      else
        for(i=3; i<wd-1; i++)
          charxy(' ',row,i);
      for(i=0; i<len; i++)
        charxy(txtBuf[i],row,textStart+i);
    }
    if(invert) toggleVideo();
  }
  if(ht>2) focus(f);
  if(wd*ht <= bufSize)
    puttext(lft,tp,lft+wd-1,tp+ht-1,buffer);
  else
    abort("not enough memory in puttext");
  return *this;
}

//

//----< alert window >-----------------------------------------

void win::alert(const char *m, int row, int col, int quit) {

  if(!alertwin) {
    alertwin = new trnWin(alertLt,alertTp,alertWd,alertHt);
    if(!alertwin)
      abort("memory allocation failure in win::alert()");
  }
  int len = strlen(m);
  alertwin->prompt("press any key to continue");
  alertwin->size(len+7,5).place(col,row).colors(WHITE,RED);
  alertwin->putText(m).paint(1);
  if(quit) exit(1);
}

//----< static alert window >----------------------------------

void win::alertMsg(const char *m,int row,int col, int quit) {

  if(!alertwin)
    alertwin = new trnWin(alertLt,alertTp,alertWd,alertHt);
  if(!alertwin)
    cout << "memory allocation failure in win::alertMsg()\n";
  alertwin->alert(m,row,col,quit);
}


//-----< is mouse postion inside win region? >-----------------

int win::myEvent(const Event &ev) {

  if(acceptAll) return 1;
  if(ev.x >= lft && ev.x <= (lft+wd -1) &&
     ev.y >= tp && ev.y <= (tp+ht-1))
    return 1;
  return 0;
}

//

//----< get window event >-------------------------------------

Event& win::getEvent(Event &ev) {

  static int lastMx=0;
  static int lastMy=0;
  ev.type = null;
  ev.x=win::mouse.wherex();
  ev.y=win::mouse.wherey();
  if(ev.x != lastMx || ev.y != lastMy) ev.type = mm;
  lastMx = ev.x;
  lastMy = ev.y;
  if(kbhit()) {
    ev.chr = getch();  ev.ext = 0;
    if(ev.chr==0) {
      ev.chr = getch(); ev.ext = 1;
    }
    ev.type = kbChar;
  }
  else if(win::mouse.lftDn())
    ev.type=mlb;
  else if(win::mouse.rgtDn())
    ev.type=mrb;
  return ev;
}

//

///////////////////////////////////////////////////////////////
//                                                           //
//  msgWin - message window definitions                      //
//                                                           //
///////////////////////////////////////////////////////////////

//----< constructor >------------------------------------------

msgWin::msgWin(int lt,int tp,int wd,int ht) : win(lt,tp,wd,ht) { }


///////////////////////////////////////////////////////////////
//                                                           //
//  trnWin - transient widnow definitions                    //
//                                                           //
///////////////////////////////////////////////////////////////

int trnWin::count = 0;

//----< constructor >------------------------------------------

trnWin::trnWin(int lt,int tp,int wd,int ht) : msgWin(lt,tp,wd,ht) {

  if(count++ == 0) {
    saveSize = wd*ht; saveBuf = new int[saveSize+1];
    if(!saveBuf) abort("memory alloc failure in trnWin()");
  }
}

//----< set window size >--------------------------------------

win& trnWin::size(int cols, int rows) {

  win::size(cols, rows);
  int tempSize=wd*ht;  int *tempBuf = new int[tempSize+1];
  if(!tempBuf)
    abort("memory allocation failure in trnWin::size()");
  int i;
  if(tempSize>=saveSize) {
    for(i=0; i<saveSize; i++)
      tempBuf[i] = saveBuf[i];
    for(i=saveSize; i<tempSize; i++)
    tempBuf[i]=(((bgc<<4)+fgc)<<8)+' ';
  }
  else
    for(i=0; i<tempSize; i++)
      tempBuf[i]=saveBuf[i];
  tempBuf[tempSize] = '\0';
  if(saveBuf) delete [] saveBuf;
  saveBuf=tempBuf; saveSize=tempSize;
  return *this;
}

//

//----< paint then hide window >-------------------------------

win& trnWin::paint(int,int,int) {

  gettext(lft,tp,lft+wd-1,tp+ht-1,saveBuf);
  prompt("press any key to continue").win::paint(1,0);
  Event ev;
  ev.type = null;
  while(ev.type == null || ev.type == mm)
    win::getEvent(ev);
  ev.type = null;
  puttext(lft,tp,lft+wd-1,tp+ht-1,saveBuf);
  return *this;
}

//----< destructor >-------------------------------------------

trnWin::~trnWin() {

  paint(1,1);
  count = 0;
  delete [] saveBuf;
}

//

///////////////////////////////////////////////////////////////
//                                                           //
//  diaWin - dialog window definitions                       //
//                                                           //
///////////////////////////////////////////////////////////////

//----< constructor >------------------------------------------

diaWin::diaWin(int lt,int tp,int wd,int ht) : win(lt,tp,wd,ht) { }

//----< send default text to dialog window >-------------------

win& diaWin::putText(const char *t) {

// reallocate text buffer for t
  int imax = strlen(t);
  if(imax >= txtSize) {
    if(txtBuf) delete [] txtBuf;
    txtBuf = new char[txtSize=wd-4];
    if(!txtBuf)
      abort("memory alloc failure in putText()");
  }
// put string at left end of buffer
  int i;
  if(imax > txtSize) imax = txtSize;
  for(i=0; i<imax; i++)
  txtBuf[i] = t[i];
  for(i=imax; i<wd-4; i++)
  txtBuf[i] = ' ';
  return *this;
}

//----< paint dialog window on terminal screen >---------------

win& diaWin::paint(int f, int, int) {

  win::paint(f, Invert, Left);

// accept input
  int row = (ht+1)>>1;
  int start = ((wd-txtSize+2)>>1);
  gotoxy(start+txtSize,row);
  editRow = row;
  editCol = start;
  return *this;
}

//----< insert char in buffer and display >--------------------

void diaWin::putChar(char ch, int col, int row) {
  charxy(ch,row,col);
  if(wd*ht <= bufSize) {
    win::mouse.hide();
    puttext(lft,tp,lft+wd-1,tp+ht-1,buffer);
    win::mouse.show();
  }
  else
    abort("not enough memory in puttext");
  gotoxy(col,row);
}

//

//----< accept user input line with edits, save in txtBuf >----

win& diaWin::process(Event &ev) {

  unsigned char ch;                      // input character

  _setcursortype(_NORMALCURSOR);
  toggleVideo();
  int i, j;
  i = textEnd(txtBuf,txtSize) + 1;     // place cursor at end
  if(i==1) i--;                        //   of text, if any
  gotoxy(editCol+i,editRow);           //   else put at origin

  while(1) {                           // collect line chars
    win::getEvent(ev);
    if(!myEvent(ev)) break;            // mouse moved outside
    if(ev.type == kbChar) {
      ch = ev.chr;
      if(ev.ext)  {   // must be extended key sequence
        switch(ch) {
          case _curLeft  : if(i>0)         i--; else putch(7); break;
          case _curRight : if(i<txtSize-1) i++; else putch(7); break;
          case _home     : i=0; break;
          case _end      : i=textEnd(txtBuf,txtSize); break;
          default        : putch(7);
        }
        gotoxy(editCol+i,editRow);
        continue;
      }
      if(ch == '\r' || ch == '\t' || ch == _esc) break;      // done
      if(i==0)
        if(isprint(ch))           // clear printable buffer contents
          for(j=0; j<txtSize; j++) {
            charxy(' ',editRow,editCol+j);
            txtBuf[j] = ' ';
          }
        else { putch(7); continue; }

      if(isprint(ch) && i<txtSize) {     // put char in buffer
        txtBuf[i] = ch;
        putChar(ch,editCol+i,editRow);
        i++; gotoxy(editCol+i,editRow);
      }
      else if(ch == _backSp) {                   // erase char
        if(i>0) {
          txtBuf[--i] = ' ';
          putChar(' ',editCol+i,editRow);
        }
      }
      else {
        putch(7); continue;                     // ring bell
      }
    }
  }
  _setcursortype(_NOCURSOR);
  toggleVideo();
  return *this;
}

//

///////////////////////////////////////////////////////////////
//                                                           //
//  itemWin - menu item window                               //
//                                                           //
///////////////////////////////////////////////////////////////

//----< constructor >------------------------------------------

itemWin::itemWin(int lt,int tp,int wd,int ht)
           : win(lt,tp,wd,ht), childCom(0) { }

//----< destructor >-------------------------------------------

itemWin::~itemWin() { delete childCom; childCom = 0;}

//

///////////////////////////////////////////////////////////////
//                                                           //
//  menuWin - window with menu pick list                     //
//                                                           //
///////////////////////////////////////////////////////////////

//----< constructor >------------------------------------------

menuWin::menuWin(int size, int lt,int tp,int wd,int ht)
  : win(lt,tp,wd,ht),child(0),next(0),max(size),current(0) {

  child = new itemWin*[size];
  if(!child)
    abort("memory alloc failure in menuWin()");
}

//----< promotion constructor >--------------------------------

menuWin::menuWin(const win& w) : win(w) { };

//----< destructor >-------------------------------------------

menuWin::~menuWin() {

  if(child) {
    int i;
    for(i=0; i<next; i++)
      delete child[i];
    delete [] child;
  }
}

//----< install menu item >------------------------------------

int menuWin::install(itemWin* wp, char *t, command *c) {

  if(next < max && next < ht-2) {
    wp->size(wd-2,1);
    wp->colors(fgc,bgc);
    wp->place(lft+1, tp+next+1);
    wp->putText(t);
    wp->regComm(c);
    child[next] = wp;
    next++;
    return 1;
  }
  return 0;
}

//----< send text to menu window >-----------------------------

win& menuWin::putText(const char* ) { return *this; }

//

//----< paint window >-----------------------------------------

win& menuWin::paint(int focus, int invert, int) {

  int i;
  win::paint(focus,0);
  for(i=0; i<next; i++)
    child[i]->paint(focus,invert);
  return *this;
}

//----< high lighter >-----------------------------------------

void menuWin::hiLite(int n) {

  if(n<0 || next <= n) return;
  child[n]->paint(0,1);
}

//----< low lighter >------------------------------------------

void menuWin::loLite(int n) {

  if(n<0 || next <= n) return;
  child[n]->paint(0,0);
}

//----< select menu item and execute its command >-------------

void menuWin::select(int n) {
  if(n<0 || next <= n) return;
  child[n]->execute();
  win::putText(child[n]->getText());
}

//

//----< select new element >-----------------------------------

void menuWin::move(int index) {

  win::mouse.hide();
  loLite(current);
  while(index <  0   ) index += next;
  while(index >= next) index -= next;
  hiLite(current=index);
  win::mouse.show();
}

//----< process window >---------------------------------------

win& menuWin::process(Event &ev) {

  hiLite(current);
  int i, newcurrent=next, last=next;
  if(win::mouse.lftDn())
    for(i=0; i<next; i++)
      if(child[i]->myEvent(ev)) { newcurrent=i; break; }
  if(newcurrent!=next) { move(newcurrent); last=current; }

  while(1) {
    win::getEvent(ev);
    if(!myEvent(ev)) break;       // mouse moved outside window
    if(ev.type==kbChar) {
      if(ev.chr==_esc || ev.chr == '\t') break;
      switch(ev.chr) {
        case  _curUp : move(current-1); break;
	case  _curDn : move(current+1); break;
        case '\r'    : select(current); return *this;
        default      : break;
      }
    }
    else if(ev.type==mlb) {
      newcurrent=next;
      for(i=0; i<next; i++)
        if(child[i]->myEvent(ev)) { newcurrent=i; break; }
      if(newcurrent==next) return *this;
      else { select(current=newcurrent); return *this; }
    }
    else {
      newcurrent=next;
      for(i=0; i<next; i++)
        if(child[i]->myEvent(ev)) { newcurrent=i; break; }
      if(newcurrent!=next && newcurrent!=last)
        move(last=newcurrent);
    }
  }
  return *this;
}

//

///////////////////////////////////////////////////////////////
//                                                           //
//  scrollWin - text viewer attached to file                 //
//                                                           //
///////////////////////////////////////////////////////////////

//----< constructor >------------------------------------------

scrollWin::scrollWin(int lt, int tp, int wd, int ht)
    : win(lt,tp,wd,ht), line(0), lineSize(160), next(0) { }

//----< destructor >-------------------------------------------

scrollWin::~scrollWin() {

  int i;
  for(i=0; i<next; i++)
    if(line[i]) farfree(line[i]);
  if(line) farfree(line);
}

//----< paint window on console screen >-----------------------

win& scrollWin::paint(int f, int, int) {

  focus(f);
  puttext(lft,tp,lft+wd-1,tp+ht-1,buffer);
  return *this;
}

//

enum {  Continue, Quit };

//----< send text to window >----------------------------------

win& scrollWin::putText(const char *file)  {

  const char *errMsg =
    "memory allocation failure in scrollWin::putText()";

  int i, j;
  ifstream in(file);
  if(!in.good()) {
    char *err = win::concat("can't open file ",file);
    alert(err,12,20,Continue);
    delete err;
    return *this;
  }
  lineSize  = 160;
  int maxLines = 250;
  if(line) {
    for(i=0; i<next; i++)
      if(line[i]) farfree(line[i]);
    farfree(line);
    line=0;
  }
  line = (char far **)farmalloc(sizeof(char far *)*maxLines);
  if(!line) alert(errMsg,12,20);
  next=0;
  while(in.good()) {
    line[next] = (char far *)farmalloc(sizeof(char) * lineSize);
    if(!line[next])
      break;
    in.getline(line[next],lineSize);
    int len = strlen(line[next]);
    for(i=len; i<lineSize; i++)
      line[next][i] = ' ';
      next++;
    if(next >= maxLines) {
      maxLines *=2;
      char far **temp;
      temp = (char far**)farmalloc(sizeof(char far*)*maxLines);
      if(!temp) break;
      for(i=0; i<next; i++)
        temp[i] = line[i];
      farfree(line);
      line = temp; temp = 0;
    }
  }
  int index = 0;
  for(j=2; j<ht; j++) {
    if(index+j-1 >= next) break;
    for(i=2; i<wd; i++)
      charxy(line[index+j-2][i-2],j,i);
  }
  return *this;
}

//

//----< process scollable text >-------------------------------

win& scrollWin::process(Event &ev) {

  int index = 0, start = 0;
  int pageInc = 2*ht/3;
  while(1) {
    win::getEvent(ev);
    if(!myEvent(ev)) break;
    if(ev.type == kbChar && ev.chr == '\t') break;
    if(ev.type == mrb) { ev.type = null; break; }
    if(ev.type == mlb ) {
      if(ev.y>=tp && ev.y<=(tp+ht/2-1)) {
        ev.type = kbChar;
        ev.chr = _pageUp;
      }
      else {
        ev.type =  kbChar;
        ev.chr = _pageDn;
      }
    }
    if(ev.type == kbChar) {
      switch(ev.chr) {
        case _curUp    : if(index > 0)         index--; break;
        case _curDn    : if(index < next-ht+1) index++; break;
        case _pageUp   : if(index > pageInc)
                           index -= pageInc;
                         else
                           index = 0;
                         break;
        case _pageDn   : if(index < next-ht+1-pageInc)
                           index += pageInc;
                         else
                           index = next-ht+1;
                         if(index<0) index=0;  // short file
                         break;
        case _ctlPgUp  : index = 0; break;
        case _ctlPgDn  : index = next-ht+1;
                         if(index<0) index=0;  // short file
                         break;
        case _curLeft  : if(start > 0)  start--; break;
        case _curRight : if(start < lineSize-wd+2) start++; break;
        case _ctlCurL  : if(start >= 20)
                           start -= 20;
                         else if(start == 0)
                           putch(7);
                         else
                           start = 0;
                         break;
        case _ctlCurR  : if(start < lineSize-wd-18)
                           start += 20;
                         else if(start == lineSize-wd-18)
                           putch(7);
                         else
                           start = lineSize-wd-18;
                         break;
        case _esc      : if(!ev.ext) ev.type=Exitw; break;
        default        : break;
      }

//

      if(ev.type==Exitw || ev.type==mrb) break;
      int i, j;
      for(j=2; j<ht; j++) {
        if(index+j-1 >= next) break;
        for(i=2; i<wd; i++)
          charxy(line[index+j-2][i+start-2],j,i);
      }
      win::mouse.hide();
      puttext(lft,tp,lft+wd-1,tp+ht-1,buffer);
      win::mouse.show();
    }
  }
  return *this;
}


//

///////////////////////////////////////////////////////////////
//                                                           //
//  winMgr - window manager definitions                      //
//                                                           //
///////////////////////////////////////////////////////////////

//----< abort with message >-----------------------------------

void winMgr::abort(char *s) {

  fprintf(stdout,"%s\n",s);
  exit(1);
}

//----< winMgr constructor >-----------------------------------

winMgr::winMgr(int size, mgrType mt) {

  if(size <= 0)
    abort("must have pos number of windows");

// save current screen params
  scroll = _wscroll;
  _wscroll = 0;
  video = directvideo;
  directvideo = 1;
  _setcursortype(_NOCURSOR);

// make window list
  next = 0;
  list = new win*[size];
  int i;
  for(i=0; i<size; i++)
    list[i] = 0;
  if(!list)
    abort("memory alloc failure in winMgr()");
  max = size;
  type = mt;
}

//----< winMgr destructor >------------------------------------

winMgr::~winMgr() {

  int i;
  for(i=0; i<next; i++)
    if(list[i]) { delete list[i]; list[i] = 0; }
  if(list) { delete [] list; list = 0; }
  max = 0;

// restore original terminal screen state
  _wscroll = 1;
  directvideo = 0;
  _setcursortype(_NORMALCURSOR);
  window(1,1,80,25);
  textbackground(LIGHTBLUE);
  textcolor(WHITE);
  clrscr();

}

//

//----< make output window >-----------------------------------

int winMgr::makeMsgWin() {
  if(next == max)
    abort("no room for new window");
  list[next] = new msgWin;
  list[next]->index(next);
  next++;
  return (next-1);
}

//----< make transient window >--------------------------------

int winMgr::makeTrnWin() {
  if(next == max)
    abort("no room for new window");
  list[next++] = new trnWin;
  return (next-1);
}

//----< make input window >------------------------------------

int winMgr::makeDiaWin() {

  if(next == max)
    abort("no room for new window");
  list[next++] = new diaWin;
  return (next-1);
}

//----< make menu window >-------------------------------------

int winMgr::makeMenuWin(menuWin &mw) {

  if(next == max)
    abort("no room for new window");
  list[next++] = &mw;
  return (next-1);
}

//----< make scrolling window >--------------------------------

int winMgr::makeScrollWin() {

  if(next == max)
    abort("no room for new window");
  list[next++] = new scrollWin;
  return (next-1);
}

//

//----< indexer >----------------------------------------------

win& winMgr::operator[](int n) {

  if(next <= 0)
    abort("no windows to access in operator[]");
  if(n<0 || next <= n) {
    fprintf(stdout,"index %d out of range in operator[]\n",n);
    return *list[0];
  }
  current = n;
  return *(list[n]);
}

//----< move window to top of display <==> bottom of list >----

int winMgr::top(int indx) {

  if(type == tiled || indx < 0 || next <= indx) return 0;
  win *save = list[indx];
  int i;
  for(i=indx; i<next-1; i++)
    list[i] = list[i+1];
  list[next-1] = save;
  for(i=indx; i<next; i++)
    list[i]->index(i);
  return 1;
}

//----< remove window >----------------------------------------

void winMgr::remove(int n) {

  if(next <= 0)
    abort("no windows to remove");
  if(n<0 || next <= n) {
    fprintf(stdout,"index %d out of range in remove()\n",n);
    return;
  }
  delete list[n];
  next--;
  int i;
  for(i=n; i<next; i++)
  list[i] = list[i+1];
  list[next] = 0;
  if(n == next)
    current = 0;
  else
    current = n;
}

//

//----< paint all windows on list >----------------------------

void winMgr::paintAll(int foc) {

  int i;
  for(i=0; i<next; i++)
    list[i]->paint(foc, 0);
}

//----< clear all windows from desktop >-----------------------

void winMgr::clearDesk() {

  window(1,1,80,25);
  textcolor(defFGC);
  textbackground(defBGC);
  clrscr();
}

//----< process window event >---------------------------------

Event& winMgr::processEvent(Event &ev) {

  if(ev.type == kbChar)
    processKbEvent(ev);
  else if(ev.type != null)
    processMouseEvent(ev);
  return ev;
}

//

//------< process window kbChar event >-------------------

Event& winMgr::processKbEvent(Event &ev) {

  if(ev.type != kbChar) return ev;
  switch (ev.chr) {
      case _esc : if(!ev.ext) ev.type = Exit; break;
      case '\t' : list[current]->paint(0);      // goto next win
                  if(current < next-1)
                    current++;
                  else
                    current = 0;
                  list[current]->paint(1);
                  break;
      default   : list[current]->process(ev);
  }
  return ev;
}

//------< process window mouse event >-------------------

Event& winMgr::processMouseEvent(Event &ev) {

  int i, newcurrent = next; // the window mouse clicked on
  if(ev.type == null) return ev;
  for(i=next-1; i>=0; i--)
    if(list[i]->myEvent(ev)) {
      newcurrent=i; break;
    }
  if(newcurrent != next) {
    win::mouse.hide();
//    if(ev.type != mm)
      list[current]->paint(0);
    if(top(newcurrent)) {
      current = next-1;
      list[current]->paint(1);
      win::mouse.show();
      list[current]->process(ev);
    }
    else
      current = newcurrent;
  }
  return ev;
}

//----< wait for keyboard or mouse event >---------------------

Event winMgr::waitForEvent(EventType et) {

  Event ev;
  ev.type = null;
  while(ev.type != et) {
    win::getEvent(ev);
    if(ev.type == null) continue;
    if(et == all) break;
    if(et == noMotion && ev.type != mm  ) break;
  }
  if(ev.type == mrb)
    exit(1);
  return ev;
}

//

#ifdef TEST_TEXTWINS

//#include "tmouse.cpp"

//----< test stub >--------------------------------------------

class testcom : public command {

  public:
    testcom(char *s);
    void execute() {
      window(1,1,80,25); gotoxy(65,24); cprintf("%s    ",teststr);
    }

  private:
    char *teststr;
};

testcom::testcom(char *s) {
  int len = strlen(s);
  teststr = new char[len+1];
  if (!teststr)
    winMgr::abort("memory allocation failure in testcom()");
  strcpy(teststr,s);
  teststr[len] = '\0';
}

//----< test routines >----------------------------------------

void main() {

  clrscr();
  winMgr window(10);
  if(!win::mouse.installed())
    winMgr::abort("mouse is not installed");
  win::mouse.show();
  win::acceptAllMouse();
  Event ev;
  int indx0 = window.makeMsgWin();
  window[indx0].prompt("window 1").putText("this is a window");
  window[indx0].BGcolor(BLACK).place(4, 2).size(30,5).paint(0);

  int indx1 = window.makeMsgWin();
  window[indx1].prompt("window 2").putText("this is another window");
  window[indx1].BGcolor(MAGENTA).place(6,4).size(30,5).paint(0);

  int indx2 = window.makeMsgWin();
  window[indx2].prompt("window 3").putText("still another window");
  window[indx2].BGcolor(CYAN).place(8,6).size(30,5).paint(0);

  int indx3 = window.makeMsgWin();
  window[indx3].size(50,5).place(14,9).colors(YELLOW,GREEN);
  window[indx3].prompt("window 4 - I'm moving");
  window[indx3].putText("press key to see another in-focus window");
  int i;
  for(i=0; i<7; i++) {
  window[indx3].place(16+2*i,13+i).paint(1);
  }
  window.waitForEvent(noMotion);

//

  window[indx3].size(40,5).prompt("window 4 - now stationary");
  window[indx3].putText("press key to repaint all windows");
  window[indx3].colors(MAGENTA,CYAN).place(35,2).paint(1);
  window.waitForEvent(noMotion);

  window.clearDesk();
  window.paintAll(0);

  int indx4 = window.makeMsgWin();
  window[indx4].place(40,5).size(34,4).colors(MAGENTA,LIGHTGRAY);
  window[indx4].putText("press any key to begin edit").paint(1);
  window.waitForEvent(noMotion);
  window.remove(indx4);

  window.clearDesk();
  window.paintAll(0);

  indx4 = window.makeDiaWin();
  window[indx4].size(50,5).place(29,6).BGcolor(MAGENTA);
  window[indx4].prompt("edit window").putText("enter here:").paint(1);
  window[indx4].process(ev).paint(0);
  window[indx3].size(50,3).prompt("data input by user");
  window[indx3].moveText(window[indx4]).place(29,3);
  window[indx3].colors(defBGC,defFGC).paint(0);

  int indx5 = window.makeMsgWin();
  window[indx5].size(40,5).putText("press any key to continue");
  window[indx5].place(39,11).paint(1);
  window.waitForEvent(noMotion);

//// here's how you write directly to the screen //////////////

  ::window(1,1,80,25);
  gotoxy(2,2);
  cout << "text from window 4 was: " << window[indx4].getText();
  gotoxy(2,3);
  cout << "text from last window was: " << window[indx5].getText();
  while(ev == EventType::null)
    getEvent(ev);

///////////////////////////////////////////////////////////////

//


// create reference to menu window built on the heap

  menuWin &menu = *(new menuWin(5));
  menu.size(17,7).place(62,16).colors(WHITE,MAGENTA);

// install menu items

  menu.install(
    new itemWin, " first item", new testcom("it's first")
  );
  menu.install(
    new itemWin, "second item", new testcom("it's second")
  );
  menu.install(
    new itemWin, " third item", new testcom("it's third")
  );
  menu.install(
    new itemWin, "fourth item", new testcom("it's fourth")
  );
  menu.install(
    new itemWin, " fifth item", new testcom("it's fifth")
  );

  int indx6 = window.makeMenuWin(menu);
  window[indx6].prompt("menu").paint(1,0);
  window[indx6].process(ev);

  int indx7 = window.makeScrollWin();
  window[indx7].place(4,10).size(60,15).colors(BLACK,YELLOW);
  window[indx7].prompt("scrollable window - esc to quit");
  window[indx7].putText("foobar.bar");
  window[indx7].putText("textwins.cpp").paint(1);
  window[indx7].process(ev).paint(0);

  window.clearDesk();
  window.paintAll(0);

  char *msg =
    "tab to cycle, mouse click to select & process, p to process";
  win::alertMsg(msg,11,2,Continue);
  win::acceptInWindow();
  while(1) {
    window.processEvent(win::getEvent(ev));
    if(ev.type == Exit || ev.type == mrb) break;
    window[indx0].prompt("sdfsf").putText( window[indx6].getText());
    window[indx0].BGcolor(BLACK).place(4, 2).size(30,5).paint(0);

  }
}

#endif
