///////////////////////////////////////////////////////////////
// tmouse.cpp - text mouse class                             //
///////////////////////////////////////////////////////////////
//                                                           //
//  tmouse.cpp - tiny text mouse class                       //
//  v1.0                                                     //
//                                                           //
///////////////////////////////////////////////////////////////
//                                                           //
//  Language:    Borland C, ver 3.1;                         //
//  Platform:    Gateway 386/25, win95                       //
//  Application: CSE687, pr3                                 //
//  Author:      Jim Fawcett, 123 Link Hall, Syracuse Univ   //
//               Syracuse, N.Y., 13244, (315)443-2654        //
//                                                           //
//           copyright (c) 1996, James W. Fawcett            //
//                all rights granted to user                 //
//                    except publication                     //
//                                                           //
///////////////////////////////////////////////////////////////


#include <dos.h>
#include <conio.h>
#include <stdlib.h>
#include <iostream.h>
#include "tmouse.hpp"

#define MOUSEDRV 0x33
union REGS regs;


//----< constructor >------------------------------------------

tmouse::tmouse() {

  if(!installed()) {
    cout << "no mouse installed\n";
    exit(1);
  }
  regs.x.ax = 7;   // set horiz range
  regs.x.cx = 1;   // min col
  regs.x.dx = 80;  // max col
  int86(MOUSEDRV, &regs, &regs);
  regs.x.ax = 8;  // set verti range
  regs.x.cx = 1;  // min row;
  regs.x.dx = 25; // max row;
  int86(MOUSEDRV, &regs, &regs);
  left = up; right = up;
}

//----< is there an installed mouse driver? >------------------

int tmouse::installed() {

  regs.x.ax = 0;
  int86(MOUSEDRV, &regs, &regs);
  return regs.x.ax;
}

//

//----< show mouse cursor >------------------------------------

void tmouse::show() {

  regs.x.ax = 1;
  int86(MOUSEDRV, &regs, &regs);
}

//----< hide mouse cursor >------------------------------------

void tmouse::hide() {

  regs.x.ax = 2;
  int86(MOUSEDRV, &regs, &regs);
}

//----< return mouse column >----------------------------------

int tmouse::wherex() {

  regs.x.ax = 3;
  int86(MOUSEDRV, &regs, &regs);
  return (1 + (regs.x.cx >> 3));
}

//----< return mouse row >-------------------------------------

int tmouse::wherey() {

  regs.x.ax = 3;
  int86(MOUSEDRV, &regs, &regs);
  return (1 + (regs.x.dx >> 3));
}

//

//----< is the left button down? >-----------------------------

int tmouse::lftDn() {

//  regs.x.ax = 3;
  regs.x.ax = 5;
  regs.x.bx = 0;
  int86(MOUSEDRV, &regs, &regs);
//  int ld = 1 & regs.x.bx;
  int ld = 1 & regs.x.ax;
  if(ld) left = down;
  return regs.x.bx;   // number of times pressed since last call
}

//----< is the right button down? >----------------------------

int tmouse::rgtDn() {

//  regs.x.ax = 3;
  regs.x.ax = 5;
  regs.x.bx = 1;
  int86(MOUSEDRV, &regs, &regs);
//  int rd = 2 & regs.x.bx;
  int rd = 2 & regs.x.ax;
  if(rd) right = down;
  return regs.x.bx;   // number of times pressed since last call
}

//----< was the left button released? >------------------------

int tmouse::lftUp() {

  regs.x.ax = 6;
  regs.x.bx = 0;
  int86(MOUSEDRV, &regs, &regs);
  int lu = 1 & regs.x.ax;
  if(lu) left = up;
  return lu;
}

//----< was the right button released? >-----------------------

int tmouse::rgtUp() {

  regs.x.ax = 6;
  regs.x.bx = 1;
  int86(MOUSEDRV, &regs, &regs);
  int ru = 2 & regs.x.ax;
  if(ru) right = up;
  return ru;
}

//

#ifdef TEST_TMOUSE

//----< test stub >--------------------------------------------

void main() {

  clrscr();
  gotoxy(2,2);
  cout << "test mouse class\n\n";

  tmouse tm;
  if(tm.installed())
    cout << "  mouse installed\n";
  else
    cout << "  no mouse installed\n";
  tm.show();

  int x, y;
  while(1) {

  // get mouse event

    gotoxy(6,6);
    cprintf("mouse position = (%d, %d) ",y=tm.wherey(),x=tm.wherex());
    gotoxy(8,8);

    if(tm.lftDn())
      cprintf("left button press  ");
    else if(tm.rgtDn())
      cprintf("right button press  ");
    else if(tm.lftUp())
      cprintf("left button release ");
    else if(tm.rgtUp())
      cprintf("right button release");

  // get keyboard event

    char ch;
    if(kbhit()) {
      ch = getch();
      switch(ch) {
        case 'h' : tm.hide(); break;
        case 's' : tm.show(); break;
        case 'c' : clrscr();  break;
        case 'q' : break;
      }
      if(ch == 'q')
        break;
    }
  }
  clrscr();
}
#endif
