#ifndef TOKS_HPP
#define TOKS_HPP
///////////////////////////////////////////////////////////////
//                                                           //
//  toks.hpp  - simple tokenizer returns words, delimiters   //
//  v1.5                                                     //
//               Manual Page                                 //
//                                                           //
//  Language:    Borland C++, ver 3.1                        //
//  System:      Gateway 386/25, DOS 6.2                     //
//  Application: CSE791 Demonstration                        //
//  Author:      Jim Fawcett, Syracuse University            //
//               123 Link Hall, (315) 443-2654               //
//                                                           //
//                    copyright (c) 1996                     //
//    all rights granted to users except right to publish    //
//                                                           //
///////////////////////////////////////////////////////////////
/*

  Module TOKS -  tokenizer treats any non-alphanumeric as a
                 delimiter.  It returns words and single
                 delimiters as tokens

  Keywords:
  =========
    token, delimiter, buffer

  TOKS operations:
  ================
  - construct a tokenizer
      ftoks toker(input);          input - existing ifstream
       toks toker(input);          input - existing istream

  - extract tokens from istream in and returns them in buffer
      int ok =  toker.getTok(buffer);

  - query tokenizer for number of lines, number of comment lines,
    number of expressions, and current brace level, respectively:
      int nl  = toker.NumLines();
      int ncl = toker.NumCLines();
      int nex = toker.NumExprs();
      int nbr = toker.NumBraces();

  - reset these statistics
      toker.resetStats();

  - display tokens
      toker.translate(buffer);     make visible string
      cout << buffer << endl;

*/
//

///////////////////////////////////////////////////////////////
//                                                           //
//               Maintenance Page                            //
//                                                           //
///////////////////////////////////////////////////////////////
//                                                           //
//  Build Process                                            //
//                                                           //
//  Files Required:                                          //
//  toks.hpp, toks.cpp, tstrs.hpp, tstrs.cpp                 //
//                                                           //
//  Integrated Environment:                                  //
//  - close any open project                                 //
//  - set Options/Compiler/Code generation/defines: to       //
//    TEST_TOKS                                              //
//  - run                                                    //
//                                                           //
//  Command Line Compiler:                                   //
//    bcc -w+ -DTEST_TOKS toks.cpp tstrs.cpp                 //
//                                                           //
///////////////////////////////////////////////////////////////
/*


  Design notes:
  =============
  - discards quoted strings and characters as well as both
    C and C++ comments
  - uses client provided buffer
  - getTok returns: 0 for end of file
                    1 for successful token
                    2 for partial token with full buffer
  - toks uses D language extensions in that it uses all str
    objects rather than char* pointers to null terminated
    arrays of characters


  History:
  ========
  v1.5 : 28 Nov 96
    - eliminated memory leak caused by ftoks construction with
      filename by elimnating constructor (not used anyway)
  v1.4 : 26 Nov 96
    - simplified getToks() by converting from buffer indexing
      to appending
    - now collect single and double quotes as tokens
    - getTok(str &buffer, int bufSize) changed to getTok(str &token)
  v1.3 : 23 Nov 96
    - declared private copy ctor, added cosmetic comments
  v1.2 : 18 Nov 96
    - replaced all C strings with str objects, made constructors
      use initialization, modified embedded stream object to
      make tokenizing strings easier.  Now have file tokenizers
      derived from string tokenizer objects.
  v1.1 : 26 Oct 96
    - made input stream a reference member
    - added constructor taking existing stream reference
    - added translation of token string to visible form
  v1.0 : 17 Oct 96
    - first release

*/
//

///////////////////////////////////////////////////////////////
//                                                           //
//               Declarations                                //
//                                                           //
///////////////////////////////////////////////////////////////

#include <fstream.h>
#include "strs.hpp"


class toks {

  public:
    toks(void);
    toks(istream &input);
    virtual ~toks() { };
    virtual int getTok(str &token);
    int good()    { return state==Good;    }
    int noExist() { return state==NoExist; }
    int eof()     { return state==Eof;     }
    int numLines ()  { return numL;   }
    int numCLines()  { return numCL;  }
    int numExprs ()  { return numExp; }
    int numBraces()  { return numBr;  }
   void resetStats() { numL = numCL = numExp = numBr = 0; }
   void translate(str &token);

  protected:
    int inComm;
    int inQuote;
    int numL, numCL, numExp, numBr;
    int getchr();
    int ungetchr(char ch);
    int eatSquote(str &token);
    int eatDquote(str &token);
    int eatCcom();
    int eatCPPcom();
    istream &in;
   enum tokState { Good, NoExist, Eof } state;
  toks& operator=(const toks &);
  toks(const toks &);
};


class ftoks : public toks {

  public:
    ftoks(ifstream &input);
    virtual ~ftoks();
};

#endif
