//////////////////////////////////////////////////////////////
//                                                          //
//   tutor.cpp    information management program, find the  //
//  v1.0       closest matches to keywords stored in the map//
//                                                          //
//                                                           //
//  Language:     Borland C++,Version 3.1                    //
//  Platform:     MS-DOS 6.2, Windows 95 v4.400.950          //
//  Application:  CSE 687  Pr3 tutor                         //
//  Author:       Mehmet SEN, CIS Grad Student at SU         //
//                (315) 443-2142, msen@top.cis.syr.edu       //
//                                                           //
//                    copyright (c) 1996                     //
//    all rights granted to users except right to publish    //
//                                                           //
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////               
#include "tutor.hpp"
#include "map1.hpp"            //map associations


//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// BASIC FUNCTIONS:  constructors, assignment, Destructor,
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/

/////////////////////////////////////////////////////////////
//  TUTOR Constructors
/////////////////////////////////////////////////////////////
inline tutor::tutor(): tutorial(map<cmpStr,str>(10)), window(7)
{}
//-------< construct with the map class >------------
inline tutor::tutor(map<cmpStr,str> & m):tutorial(m) , window(7)
{
    if(!win::mouse.installed())
	winMgr::abort("mouse is not installed");
    win::mouse.show();
    win::acceptAllMouse();
}
/////////////////////////////////////////////////////////////
//  TUTOR COPY Constructor
/////////////////////////////////////////////////////////////
tutor::tutor(const tutor & t):tutorial(t.tutorial) , window(t.window)
{
    ev=t.ev;
    database=database;
    valW=t.valW;
    keyW=t.valW;
    senW=t.senW; //sentence window
    aleW=t.aleW;
}
/////////////////////////////////////////////////////////////
//  TUTOR ASSIGNMENT OPERATOR
/////////////////////////////////////////////////////////////
tutor & tutor::operator=(const tutor &t) {
    if (this==&t) return *this;
    window=t.window;
    ev=t.ev;
    tutorial=t.tutorial;
    database=database;
    valW=t.valW;
    keyW=t.valW;
    senW=t.senW; //sentence window
    aleW=t.aleW;
    return *this;
}

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// SIDE SERVICE FUNCTIONS; defaults , abort , space allocation
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/

/////////////////////////////////////////////////////////////
// Terminate the program with the error message
/////////////////////////////////////////////////////////////
template <class key, class val>
void  map<key,val>::abort(const char *s){
	cout <<s<<endl;
	exit(1);
}
/////////////////////////////////////////////////////////////
// initialize the windows for the user interaction
/////////////////////////////////////////////////////////////
void tutor::openWin(){
    keyW = window.makeMsgWin();
    valW = window.makeScrollWin();
    senW = window.makeDiaWin();
    aleW = window.makeScrollWin();
    //clrscr();
}
/////////////////////////////////////////////////////////////
//  Display a specific message
/////////////////////////////////////////////////////////////
void tutor::showAlert(const char * str,int x,int y) {
      /*	ofstream current("alert.dat",ios::out);
	current << str <<endl;  current.close();
	window[aleW].place(5,5).size(60,15).colors(WHITE,YELLOW);
	window[aleW].prompt("press any key to continue");
	window[aleW].putText("alert.dat").paint(0);
	window[aleW].putText("alert.dat").paint(0);
	window.waitForEvent(noMotion);
	win::getEvent(ev);
	if (ev.chr == 27 || ev.type == mrb) ;

      */	win::alertMsg(str,10+y,2+x,0);
}
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// MAIN SERVICE FUNCTIONS
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/

/////////////////////////////////////////////////////////////
// Load the tutorial database from the file
/////////////////////////////////////////////////////////////
void tutor::load(char * filename,map<cmpStr,str> &tomap) {
 //   database=filename;
    ifstream in(filename);
    cmpStr key;
    str val;
    if(!in.good()) {
	cout << "  can't open file. " << filename << endl;
	return;
    }
    ftoks toker(in);
    int notDone,found;
    cmpStr token(70);
    do {
	  notDone = toker.getTok(token);
	  if ( token == "<" ) {
		found =1;
		notDone = toker.getTok(key);
		val="";
	  } else if ( token == ">" ) {
		tomap.insert(key,val);
		found=0;
	  } else if ( found )	val += token;
    } while(notDone);
//    tutorial.dump(cout);
}
/////////////////////////////////////////////////////////////
// Runs in an interactive mode; tokenize sentence, find best matched key
// and additional key list, show best mached key value and the list. Ask for
// another sentence until user presses to the esc button
/////////////////////////////////////////////////////////////
void tutor::run(cmpStr sentence){
    #define TOKSIZE  20
    openWin();
    window.clearDesk();
    showAlert(" `_` between words means exact matching");
    showAlert(" Press ESC to abort the program at any time",2,2);
    while (1) {
	map<int,cmpStr> bests;
	cmpStr token(TOKSIZE);
	istrstream in((char*)sentence.cStr());
	toks stoker(in);
	while (stoker.getTok(token))
	       if(token.len() > 1)
			getTutorial(token,bests);
	showValue(bests.quickValue(0));
	showKeys(bests);
	showValue(bests.quickValue(0));
	if (!getSentence(sentence)) break;
    }
}
/////////////////////////////////////////////////////////////
// Get another sentence from the user
/////////////////////////////////////////////////////////////
int tutor::getSentence(cmpStr & text) {
    window[senW].size(80,5).place(1,21).BGcolor(MAGENTA);
    window[senW].prompt("Enter new sentence").putText("").paint(1);
    //  window.processEvent(win::getEvent(ev));
    window[senW].process(ev).paint(0);
    text=window[senW].getText();
    if (ev.chr == 27 || ev.type == mrb) return 0;
    return 1;
    /*   char inp[255];  cin >> inp;  text=inp;  cout << text;  */
}
/////////////////////////////////////////////////////////////
// Display the key list  to the console
/////////////////////////////////////////////////////////////
void tutor::showKeys(const map<int,cmpStr> & bests){
	str keys;
	keys="";
	for (int i=0; i< bests.length();i++)  {
		keys +=  bests.quickValue(i);
		keys += " ";
	}
	window[keyW].size(80,5).place(1,16).colors(YELLOW,BLACK);
	window[keyW].prompt("Best Matches");
	window[keyW].putText(keys.cStr()).paint(0);
	window[senW].paint(0);
}
/////////////////////////////////////////////////////////////
// Display the best matched value to the screen, also save it to
// the file as a log info
/////////////////////////////////////////////////////////////
void tutor::showValue(const cmpStr & key) {
	str val =" ";
	val += (tutorial[key]);
	ofstream current("last.dat",ios::out);
	current << key<<endl<<val;	    //	best.dump(current);
	current.close();
	window[valW].place(1,1).size(80,15).colors(BLACK,CYAN);
	window[valW].prompt("BEST MATCH VALUE");
	window[valW].putText("last.dat").paint(0);
//	window[valW].process(ev).paint(0);
}
/////////////////////////////////////////////////////////////
// add the key  to the matched list; points show how exact it is
/////////////////////////////////////////////////////////////
void tutor::addList(int points, cmpStr keyval, map<int,cmpStr> & bests) {
	int i;
	while (bests.has(points,&i))
		points++;
	bests.insert(points,keyval);
}
/////////////////////////////////////////////////////////////
// Search the tutorial for the specific key, i.e. map search
/////////////////////////////////////////////////////////////
void tutor::getTutorial(cmpStr key, map<int,cmpStr> & bests) {
	int i,points;
	str value;
	value=tutorial.search(key,&i,&points);
	if ( points != -1 )
		addList(points, tutorial.checkKey(i),bests);
}

#ifdef TEST_TUTOR
/////////////////////////////////////////////////////////////
// test stub for the tutor module
/////////////////////////////////////////////////////////////
main(int argc, char *argv[])
{
    map<cmpStr,str>   bm;
    tutor::load("default.dat",bm);
    map1<cmpStr,str>  derived_m(bm);    //construct a derived map
    // I am aware of duplications of the map associations by copy const.
    //this can be avoided by calling a function to load the derived map by
    //assignment in it, and destructing the base map at return,but because
    //of READABILITY reasons I am leaving it in that way.
    tutor t(derived_m);              // construct map with the derived map class
    cmpStr sentence="";      // the first sentence
    for (int i=1;i<argc;i++) {// from the argument list
	sentence+=argv[i];
	sentence+=" ";
    }
//    t.load("default.dat");     // load the default tutorial info
    t.run(sentence);		// run the interactiveturorial   mode
    return 1;
}
#endif

//////