#ifndef TUTOR_HPP
#define TUTOR_HPP


///////////////////////////////////////////////////////////////
//                                                           //
//  tutor.hpp    information management program, find the    //
//  v1.0         closest matches to keywords stored in the map//
//                                                           //
//                manual page                                //
//                                                           //
//  Language:     Borland C++,Version 3.1                    //
//  Platform:                                                //
//                MS-DOS 6.2, Windows 95 v4.400.950          //
//  Application:  CSE 687  Pr3 tutor                         //
//  Author:       Mehmet SEN, CIS Grad Student at SU         //
//                (315) 443-2142, msen@top.cis.syr.edu       //
//                                                           //
//                    copyright (c) 1996                     //
//    all rights granted to users except right to publish    //
//                                                           //
///////////////////////////////////////////////////////////////
//    information about the class operations
/*
	Tutor class
	===========================================
	Manage the accesses to the information stored in the map class
	It accepts user requests for the information, parses requests
	into tokens and finds the closest match to keywords stored in a
	map. It then displays text associated with the keyword, like a
	dictionary or help system. It also lists other related keys
	which then may be selected for another matching.

	constructors
	===========================================
	tutor t;		void constructor
	tutor t(map);           construct t with a map
	tutor t1(t);		copy constructor

	operations
	===========================================
        t1=t2			assignment operation
	load(filename,map)	loads the map of the tutor from the file
	run (sentence)		run tutorial with the sentence,(
				repeats by asking new sentences)
	getSentence(cmpStr &)   read another sentence
	showKeys(bests_map)     lists the keys from the bests map
	showValue(key)   	display the best  key value
	showAlert(str)          alert a message
	addList(points,keyval,bests) insert the matched key to the bests list
	getTutorial(key,bests) 	find the best matched key list for the key
				from the map
	openWin()		initialize the windows for the display
	~tutor() 		destruct the tutor

	hidden
	======
	abort			abort the system

*/

//
///////////////////////////////////////////////////////////////
//                                                           //
//  Build Process:                                           //
//                                                           //
//  Files required for building test stub: ()                //
//      map.hpp, map.cpp, cmpStrs.hpp, cmpStrs.cpp           //
//      map1.hpp, map1.cpp,strs.hpp, strs.cpp                //
//	textwins.hpp, textwins.cpp, tmouse.hpp, tmouse.cpp   //
// 	tutor.hpp tutor.cpp, toks.hpp, toks.cpp 	     //
//							     //
//  Build from integrated environment:                       //
//  - close any open project                                 //
//  - set Options\Compiler\Code generated\defines to         //
//    if you want to perform general tests on string classes //
//      TEST_TUTOR                                           //
//    if you want to test it to list the files in current dir//
//  - open tutor.cpp in edit window and run                  //
//                                                           //
//  - command line build,                                    //
//      bcc -w+ -DTEST_TUTOR textwins.cpp tmouse.cpp         //
//                          tutor.cpp map.cpp map1.cpp       //
//			toks.cpp  cmpstrs.cpp strs.cpp       //
//                                                           //
///////////////////////////////////////////////////////////////
/*


	maintenance history
	===========================================
	- v1.0 :	March 26,1997
			-first release
	design notes
	===========================================



		     +---+----+    +---+----+
		     | cmpStr |>---| str    |
		     +---+----+    +---+----+
			 |             |
			 |   +---------+
			/\  /\
			\/  \/
    +-------+        +--+---+-+      +-------+
    |  map  |*<----<>| TUTOR  |<>----| toks  |
    +---+---+        +---+--+-+      +-------+
	|                /\ /\
       / \               \/ \/
    +---+----+           |  |
    | map1   |           |  |
    +--------+           |  |
			 |  |
			 |  |
			 |  +---------------------------+
			 +--------------+               |
					|               |
					|               |
					|               |
		  +-------+        +----+---+      +----+--+
		  |  win  |o<----<>| winMgr |<>--->| Event |
		  +---+---+        +--------+      +-------+
		      |
		     / \
     +-----------+------------+-----------------+-------------+
     |           |            |                 |             |
 +---+----+  +---+----+  +----+----+       +----+----+  +-----+-----+
 | msgWin |  | diaWin |  | itemWin |o<---<>| menuWin |  | scrollWin |
 +--------+  +--------+  +---------+       +----+----+  +-----------+
     |                       < >
     |                        |
    / \                       |
 +--------+  +---------+      |
 | trnWin |  | command |o<----+
 +--------+  +----+----+
		  |
		 / \
		  |
	     +----+----+
	     | testCom |
	     +---------+

	Tutor class use a map class to store information, However, because
	of the virtual functions in the map class, tutor class can be cons-
	tructed by any derived map class with special near match functions.
	You will see the usage of this in the tutor test stub, where tutor
	object is constructed with the map1 class. This loosely coupling
	provides the users improve new information retrieval systems without
	touching the tutor module.

	Infact, instead of making map1 module, user can use a small derived
	map class only with two functions in the tutor.cpp file.

	The sentences of the user are tokenized by using toks module.
	Also, the map class key-value associations are loaded from the file
	by using toks module.

	Tutor class use an window manager to interact with the user.

	Map class keeps key and value associations.

*/
//
///////////////////////////////////////////////////////////////               
//                                                           //               
//                      declarations                         //               
//                                                           //               
///////////////////////////////////////////////////////////////               
#include <stdlib.h>
#include <iostream.h>
#include <strstream.h>
#include "textwins.hpp"       //display
#include "map.hpp"            //map associations
#include "toks.hpp"	      //tokenize the strings
#include "cmpstrs.hpp"
#include "strs.hpp"

class tutor {
public:

    tutor();
    tutor(map<cmpStr,str> &m);
    tutor(const tutor &t);			// copy constructor
    tutor & operator=(const tutor &t); //assign.

    static void load(char * filename,map<cmpStr,str> & tomap);//load information
    void run(cmpStr sentence);          //interact with user
    int getSentence(cmpStr &);
    void showKeys(const map<int,cmpStr> & bests);
    void showValue(const cmpStr & key);
    void showAlert(const char * str,int x=0,int y=0);
    void addList(int points, cmpStr keyval, map<int,cmpStr> & bests);
    void getTutorial(cmpStr,map<int,cmpStr> & bests);
    void openWin();
    virtual ~tutor() {  }

private:
    void abort(const char *s);
    winMgr  window;
    Event ev;
    map<cmpStr,str> & tutorial;
    cmpStr database;
    int valW,keyW,senW,aleW;          //value,key,sentence windows
};



#endif // TUTOR_HPP

